#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include "playback.h"

#define VSAAS_PLAYBACK_CHANNEL 1

void PrintErrHandling(int nErr);

extern AV_Client gClientInfo[MAX_CLIENT_NUMBER];
extern int ExPasswordAuthCallBackFn(const char *account, char *pwd, unsigned int pwd_buf_size);
extern int ExTokenAuthCallBackFn(const char *identity, char *token, unsigned int token_buf_size);
char gInfoFilePath[] = "./frames.info", gVideoFilePath[] = "./frames.bin";

int GetFrameInfo(FILE *info_file, ENUM_FRAMEFLAG *frame_type, unsigned int *frame_size, unsigned int *last_frame)
{
	char line_buf[1024] = {0};
	char type;
	int pos;
	char *ret = NULL;

	ret = fgets(line_buf, 512 + 1, info_file);
	if (ret == NULL)
	{
		return -1;
	}

	sscanf(line_buf, "%c %d %d", &type, &pos, frame_size);
	*frame_type = type == 'I' ? IPC_FRAME_FLAG_IFRAME : IPC_FRAME_FLAG_PBFRAME;
	
	if(feof(info_file)){
	    *last_frame = 1;		
	}

	return 0;
}

void *thread_PlayBack(void *arg)
{
    FILE *info_file, *video_file;
    char frame_buf[VIDEO_BUF_SIZE] = {0};
    char info_buf[1024] = {0};
    int ret, interval_ms;
    unsigned int last_frame = 0, frame_ts_ms = 0, frame_size;
    ENUM_FRAMEFLAG frame_type;
	FRAMEINFO_t frameInfo;
    char fps_symbol[10];
    int fps;

	int lock_ret;
	int SID = *((int *)arg);
    free(arg);
    AV_Client *p = &gClientInfo[SID];

    AVServStartInConfig avStartInConfig;
    AVServStartOutConfig avStartOutConfig;

    memset(&avStartInConfig, 0, sizeof(AVServStartInConfig));
    avStartInConfig.cb               = sizeof(AVServStartInConfig);
    avStartInConfig.iotc_session_id  = SID;
    avStartInConfig.iotc_channel_id  = p->playBackCh;
    avStartInConfig.timeout_sec      = 30;
    avStartInConfig.password_auth    = &ExPasswordAuthCallBackFn;
    avStartInConfig.token_auth       = &ExTokenAuthCallBackFn;
    avStartInConfig.server_type      = SERVTYPE_STREAM_SERVER;
    avStartInConfig.resend           = ENABLE_RESEND;
#if ENABLE_DTLS
    avStartInConfig.security_mode = AV_SECURITY_DTLS; // Enable DTLS, otherwise use AV_SECURITY_SIMPLE
#else
    avStartInConfig.security_mode = AV_SECURITY_SIMPLE;
#endif
    printf("playback channel , avStartInConfig.iotc_channel_id : %d\n", avStartInConfig.iotc_channel_id);
    avStartOutConfig.cb              = sizeof(AVServStartOutConfig);
    int avIndex = avServStartEx(&avStartInConfig, &avStartOutConfig);
    
    if(avIndex < 0)
    {
        printf("avServStart2 failed SID[%d] code[%d]!!!\n", SID, avIndex);
        pthread_exit(0);
    }
    printf("[%s:%d] SID[%d] avIndex[%d] start\n", __FUNCTION__, __LINE__, SID, avIndex);

    info_file = fopen(gInfoFilePath, "r");
    video_file = fopen(gVideoFilePath, "r");
    if(info_file == NULL || video_file == NULL){
	printf("open info_file fail or open video_file fail\n");
    }

    //get FPS from video info file
    char *ret_str = fgets(info_buf, 512, info_file);
    if (ret_str == NULL)
    {
            return NULL;
    }

    sscanf(info_buf, "%s %d", fps_symbol, &fps);
    printf("%s() %s=%d\n", __FUNCTION__, fps_symbol, fps);
    interval_ms = 1000 / fps;
    /*
     *  search play back by p->playRecord.stTimeDay
     */
	while(p->bEnableRecordVideo)
	{
		ret = GetFrameInfo(info_file, &frame_type, &frame_size, &last_frame);
		if (ret < 0)
		{
			printf("%s() no more record frame,close session[%d]\n", __FUNCTION__, SID);
			avServStop(avIndex);
			IOTC_Session_Close(SID);
			break;
		}else if(frame_size <= 0){
			continue;
		}	
        
		printf("%s() avIndex=%d frame_type=%d frame_ts_ms=%d frame_size=%d, last_frame=%d\n", __FUNCTION__, avIndex, frame_type, frame_ts_ms, frame_size, last_frame);
		ret = fread(frame_buf, 1, frame_size, video_file);

		memset(&frameInfo, 0, sizeof(FRAMEINFO_t));
		frameInfo.codec_id = MEDIA_CODEC_VIDEO_H264;
		frameInfo.flags = frame_type;
		frameInfo.timestamp = frame_ts_ms;
		
		if(last_frame == 1){
		    printf("last frames\n");
		    frameInfo.tags = 1;	
		}

		ret = avSendFrameData(avIndex, frame_buf, frame_size, &frameInfo, sizeof(FRAMEINFO_t));

		if (ret < 0)
		{
			printf("avSendFrameData Error\n"); 
			PrintErrHandling(ret);
			break;
		}

		usleep(20000);

		frame_ts_ms += interval_ms;
	}


	avServStop(avIndex);
	fclose(info_file);
	fclose(video_file);
	lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
	if(lock_ret)
		printf("Acquire SID %d rwlock error, ret = %d\n", SID, lock_ret);
	gClientInfo[SID].playBackCh = -1;
	lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);
	printf("[%s:%d] SID[%d] avIndex[%d] exit\n", __FUNCTION__, __LINE__, SID, avIndex);
	pthread_exit(0);

}

int Handle_IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL(int sid, int av_index, char* req_buf)
{
    printf("%s called, sid[%d], av_index[%d]\n", __func__, sid, av_index);

    int *tmp_sid = NULL, ret = 0, lock_ret = 0;
    pthread_t ThreadID;
    SMsgAVIoctrlPlayRecord *req = (SMsgAVIoctrlPlayRecord *)req_buf;
    SMsgAVIoctrlPlayRecordResp resp;

    printf("[%s:%d] av_index[%d] Receive IOCTRL command: IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL cmd[%d]\n", __FUNCTION__, __LINE__, av_index, req->command);
    printf("stTimeDay[%04d:%02d:%02d-%02d:%02d:%02d]\n", req->stTimeDay.year, req->stTimeDay.month, req->stTimeDay.day,
                                                         req->stTimeDay.hour, req->stTimeDay.minute, req->stTimeDay.second);


    if(req->command == AVIOCTRL_RECORD_PLAY_START)
    {
        printf("AVIOCTRL_RECORD_PLAY_START \n");
        memcpy(&gClientInfo[sid].playRecord, req, sizeof(SMsgAVIoctrlPlayRecord));
        //get writer lock
        lock_ret = pthread_rwlock_wrlock(&gClientInfo[sid].sLock);
        if(lock_ret)
            printf("Acquire SID %d rwlock error, ret = %d\n", sid, lock_ret);

        resp.command = AVIOCTRL_RECORD_PLAY_START;
        printf("playback on playBackCh[%d]\n",gClientInfo[sid].playBackCh);
        if(gClientInfo[sid].playBackCh < 0)
        {
            gClientInfo[sid].bEnableRecordVideo = 1;
            gClientInfo[sid].playBackCh = VSAAS_PLAYBACK_CHANNEL;
            printf("gClientInfo[sid].playBackCh : %d\n", gClientInfo[sid].playBackCh);
            resp.result = gClientInfo[sid].playBackCh;
        }
        else
        {
            resp.result = -1;
        }

        //release lock
        lock_ret = pthread_rwlock_unlock(&gClientInfo[sid].sLock);
        if(lock_ret)
            printf("Release SID %d rwlock error, ret = %d\n", sid, lock_ret);

        if(resp.result >= 0)
        {
            printf("start playback result[%d]\n", resp.result);

            tmp_sid = (int *)malloc(sizeof(int));
            *tmp_sid = sid;
            if((ret = pthread_create(&ThreadID, NULL, &thread_PlayBack, (void *)tmp_sid)))
            {
                printf("pthread_create thread_PlayBack ret[%d]\n", ret);
                resp.result = -1;
            }
            else
            {
                pthread_detach(ThreadID);
            }
        }
        else
        {
            printf("Playback on SID %d is still functioning\n", sid);
        }

        printf("Sending resp.result[%d]\n", resp.result);
        if(avSendIOCtrl(av_index, IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL_RESP, (char *)&resp, sizeof(SMsgAVIoctrlPlayRecordResp)) < 0)
            return -1;
    }
    else if(req->command == AVIOCTRL_RECORD_PLAY_STOP)
    {
        printf("AVIOCTRL_RECORD_PLAY_STOP \n");
        //get writer lock
        lock_ret = pthread_rwlock_wrlock(&gClientInfo[sid].sLock);
        if(lock_ret)
            printf("Acquire SID %d rwlock error, ret = %d\n", sid, lock_ret);
        gClientInfo[sid].bEnableRecordVideo = 0;
        //release lock
        lock_ret = pthread_rwlock_unlock(&gClientInfo[sid].sLock);
        if(lock_ret)
            printf("Release SID %d rwlock error, ret = %d\n", sid, lock_ret);
    }

    return 0;
}

